       1  #!/usr/bin/perl
       2  #
       3  # CLARIFY OPTIONS
       4  #	output format:
       5  #	- verbose
       6  #	- sort by (file size), (file size * (# files of that size -1)
       7  #	- sorting order: largest first, smallest first, no sorting
       8  #	- human readable
       9  #	- linked files: list ALL, FIRST, NONE
      10  #
      11  #
      12  #	Tue Nov 27 13:53:54 EST 2001
      13  # this is dup7
      14  # ??? better names: broom: helps you clean up / sweep up files
      15  
      16  # Thu Jan  9 06:54:03 EST 2003
      17  # !!! make the names a bit better, explain how the parts work together
      18  #
      19  # OVERVIEW:
      20  # Given a list of regular files, this program determines what files are identical
      21  # by content, regardless of file name.
      22  #
      23  # The main function groups files by filesize and compares all those files to each other.
      24  # There are 2 filters to help for a "large" (*1) number of files
      25  # "filterByFirstBlock"
      26  #	in: a list of filenames
      27  #	out: a zzz LoL of filenames that match by first block
      28  #		(several lists may result, not just one)
      29  #	zzz I need a generic standard name for LoL
      30  #
      31  #	The first block of each file is read into an array which is used for all further comparisions,
      32  #	thus minimizing disk access.
      33  #	For files of size <= the blocksize, this is all you need!
      34  #	For files > the blocksize, further testing is needed.
      35  #
      36  # "filterBymd5"
      37  #	in: a list of filenames
      38  #	out: a zzz LoL of filenames that match by md5digest
      39  #		(several lists may result, not just one)
      40  #
      41  #	The md5sum is calculated for each file.
      42  #	Lists are made of all files with matching md5sum, indicating that they're PROBABLY identical.
      43  #	While md5sum is really REALLY good, it's not perfect,
      44  #	so for "belt and suspenders" assurance, it's best to fully compare the files too.
      45  #	zzz need an option to use md5sum only, no further testing
      46  #
      47  #	This phase has the ADVANTAGE that each file is read only once.
      48  #	This phase has the DRAWBACK that each file is read entirely from end to end
      49  #	even if they differ at the beginning.
      50  
      51  
      52  # DETAILS:
      53  
      54  # (*1) what constitutes a "large" number of files?
      55  # it's calculated via zzz, altered by zzz options/args/variables
      56  
      57  
      58  
      59  # TO DO:
      60  # add: getopt handling
      61  # -v: verbose (the jsjv stuff)
      62  # -m: machine readable output (default is human-friendly output)
      63  # -z: don't report zero length files
      64  # -s: don't report SYMLINKS
      65  #     (unless part of a non-symlink match? or only one match?)
      66  # -gnnn: only files greater than size n !!! GOTTA ALLOW k, m, ...
      67  #
      68  # -f: FORCE MD5sum only (no file compare)  WARNING: may be wrong
      69  # -m: use MD5sum even for 2 files (set low threshold/parameters)
      70  # -x: save MD5sum results in a format for md5sum verify/check
      71  # -ziii: set parameters for when to use MD5sum: file size, # files, ...
      72  #	perhaps allow an entire equation via 'eval'
      73  # -biii: set block size
      74  #
      75  # -c: FORCE file compare (no md5sum)
      76  #
      77  #
      78  # ADD: other ways to read the list-of-files: STDIN, filename, other command-to-run
      79  #	(ARG or ENV VAR: command-and-args to run for "find")
      80  # REASONS TO USE the external command "find" instead of a Perl built-in
      81  # 1) less code to write and maintain, particularly for search parameters such as "xdev"
      82  # 2) eliminates inconsistencies of "this find does THIS and that find generates THAT"
      83  # 3) lets you run the command by itself to verify/refine the search
      84  # 4) it's in the original Unix spirit that software tools ought to be
      85  #	ORTHOGONAL (not duplicate functionality),
      86  #	do ONE THING WELL and pipeline them together as needed
      87  #
      88  # REASONS TO USE a Perl built-in "find"
      89  # - makes a stand-alone executable for non-Linux environments
      90  #
      91  # I NEED HELP:
      92  # on non Unix/Linux systems:
      93  # a) how to detect duplicate shortcuts or other ways to give
      94  #	one file multiple names
      95  # b) can it auto-detect that or must it be args/env variable?
      96  #
      97  
      98  # BETTER SYMLINK HANDLING
      99  # - filter out all HARD LINKS and SYMLINKS at the beginning
     100  #	? make a hash-of-strings?
     101  #	ex: a1, a2, a3, a4 all symlinks
     102  #	a) match all files by file size
     103  #	b) for all files of the same size
     104  #		links{"a1"} -> a2, a3, a4
     105  #		just "a1" in the main list
     106  #		(verify it's REGULAR FILE since stat() was done)
     107  #	c) when printing results, select
     108  #		only files matching (no hard/sym links)
     109  #		only hard/sym links (no other matches)
     110  
     111  # YET ANOTHER SORTING CRITERION:
     112  # - by file size  <<< can be done in advance
     113  # - by total space: (file size) * (# duplicates)  <<< can NOT be done in advance
     114  #	those will save the most by linking/eliminating duplicates
     115  
     116  
     117  # use strict;  ### gotta solve things first
     118  use File::Compare;
     119  use Digest::MD5;
     120  use Getopt::Std;
     121  
     122  # _____________________________________________________________________
     123  # consider: pass to filter out all entries of <n files (normally 2)
     124  # so no more "skipping entry ...",
     125  # the total count is more useful then!
     126  # NEEDS AN OPTION: prune entries with <n entries of the same size
     127  #	(default is 2)
     128  # NEEDS AN OPTION: prune entries <n bytes in size
     129  #	(default is 0???)
     130  # IN: hash of arrays, by file size
     131  # OUT: hash of arrays, only with >n entries, only >n filesize
     132  #	# of files total
     133  #	# of entries
     134  #
     135  
     136  # _____________________________________________________________________
     137  
     138  # POSSIBLE OPTIMIZE: sort out HARD LINKS before comparing
     139  #	no need to diff/cmp the hard links!,
     140  #	but SKIP THAT IF IT'S NOT SUPPORTED: nfs / win-98 / ...
     141  # OPTION: hard links: list them or don't list them
     142  # _____________________________________________________________________
     143  
     144  # will this work?
     145  # for all files of the same size
     146  #     if (hard link detection desired)
     147  #	yank all with the same hard links
     148  #       if (print hard links) print the list
     149  # continue with dup file detection: first block compare, md5, cmp
     150  # HOW CAN I STILL GENERATE THE TOTAL:
     151  #	5 identical files of size 11111  <<< THIS COUNT
     152  #	  2 are LINKED files:
     153  #
     154  # IDEA: have a pass like compare_by_fc (), returning a list-of-lists
     155  
     156  # _____________________________________________________________________
     157  
     158  # TO DOCUMENT: 1) how to use gnu find, other finds
     159  # 2) with options like: prune out /$HOME/.netscape
     160  # _____________________________________________________________________
     161  
     162  #
     163  # OPTION: -f 0: format 0/1/2/3/...
     164  
     165  # format 0: human readable
     166  
     167  # format 1:
     168  #   S file-size                                <<< may be suppressed separately
     169  #   L link-file-1 link-file-2 ... link-file-n  <<< may be suppressed
     170  #   D dup-file-1 dup-file-2 ... dup-file-n
     171  
     172  # jsj ; TO DO:
     173  # ARGS: block size, verbose, version
     174  # ARG: post-process filenames to escape shell-sensitive characters
     175  #	$_ =~ s/([ \$&'"(])/\\$1/g;  # escape strange characters
     176  
     177  
     178  #	require 'getopts.pl';
     179  #	&Getopts('s:x');
     180  #	if (defined($opt_s))
     181  #	elsif ($opt_x)
     182  # THAT IS OBSOLETE:
     183  # Suggested alternatives: Getopt::Long  or  Getopt::Std
     184  
     185  # SKIP md5 under certain circumstances
     186  # ??? perhaps some args to change those rules?
     187  
     188  # ARG: alternat output formats for easier SHELL/PERL reading
     189  #	option: list/suppress symbolic links
     190  
     191  
     192  
     193  # DOCUMENT: the algorithms, performance
     194  # ASK FOR HELP: when to use md5, when to skip
     195  # (can that be determined by number of files and their length, or are
     196  # other things needed to be known?)
     197  
     198  # !!! need: documentation style for comments
     199  
     200  # Author: Jeffrey Jonas  jeffj@panix.com
     201  # version 1.0  March 2001
     202  
     203  # EMPHASIZE:
     204  # this only lists the files - it's up to you how to eliminate the duplicates.
     205  #	jsj mention the rm / link solution
     206  # files are sorted with largest file size first since those are the ones
     207  #	to handle first!
     208  
     209  # hard links are identified since they don't take additional disk space
     210  
     211  
     212  # can skip small files with FIND options -size +nnn
     213  
     214  # this does NOT handle:
     215  # 1) soft links.
     216  #	it's deliberately "find ... -type f" since I can't think of anything
     217  #	useful to do with symbolic links.
     218  #	Symbolic links are *NOT* followed since they may point to directories
     219  #	and cause loops.
     220  # 2) directories
     221  # 3) files that are similar
     222  #	ex: "This is a file" to "This is a file / with more".
     223  #	It's hard enough finding files that are identical.
     224  #	I'm unsure how to score a diff for being "similar enough".
     225  
     226  
     227  # jsj DESCRIBE THE ALGORTIHMS AND HOW THIS EVOLVED:
     228  
     229  # there are 2 main methods of finding duplicate files
     230  # compare the files themselves, or compare their checksums.
     231  
     232  # comparing the files directly works well IF
     233  # 1) the files really match, since the files MUST be compared to assure
     234  #	they're really identical
     235  # 2) the files don't match within the first block or 2
     236  
     237  # The worst case scenario: files mis-match only near the end,
     238  # so every file is read in entirely for every attempt at a match.
     239  # for 3 files, that's 3 compares (a,b) (a,c) (b,c),
     240  # so files are read end to end 6 times.
     241  # for 4 files, that's 6 compares (a,b) (a,c) (a,d) (b,c) (b,d) (c,d)
     242  # so files are read end to end 12 times.
     243  
     244  # jsj: general formula for (n) files
     245  
     246  
     247  # jsj MD5
     248  # best case: no files match: so each file is read end to end only once
     249  # to generate the checksum.
     250  #
     251  # The worst case scenario: many files match, since they need to then be
     252  # re-read in entirety to verify every potential match with cmp()
     253  
     254  
     255  
     256  
     257  # credits:
     258  # jsj to the md5 fellow
     259  
     260  # to Chip Unicorn for cutting the performance gordion knot
     261  # and suggesting the pass with just reading the first block.
     262  
     263  # !!! need: arg parsing stuff
     264  
     265  my $usage = "NO ARGS!\
     266  This perl script finds all duplicate files by content, not file name\
     267  usage: $0 <options to find>\
     268  example: $0 . -mount\n\n";
     269  die $usage if ($#ARGV < 0);
     270  
     271  # jsj: give example with -size +2k
     272  # jsj: give example with -prune
     273  
     274  # jsj: give example like
     275  #	$HOME ~friend -size +100k -mount
     276  
     277  #
     278  
     279  our($opt_v);
     280  getopts('v');
     281  
     282  my ($verbose1, $verbose2, $verbose3);
     283  
     284  $verbose1++ if ($opt_v);
     285  my %names_by_size;
     286  
     287  # Form the external "find" command using the arg list.
     288  # Gather filenames by size using a hash of lists
     289  # names_by_size { file_size } -> (name1, name2, name3, ...)
     290  
     291  # jsj !!! better error handling if FIND fails:
     292  # - get the retc
     293  # - just let STDERR go thru?
     294  #
     295  # jsj: perhaps the "-type f" needs supression at times?
     296  
     297  print STDERR "jsjv: find is generating the list of files\n"
     298      if ($verbose1);
     299  
     300  my $cmd = "find " . join (" ", @ARGV) . " -type f -print |";
     301  # print "cmd = ", $cmd, "\n";
     302  
     303  open (pipefd, $cmd);
     304  
     305  my ($filecount) = 0;  # count just for stats
     306  while (<pipefd>)
     307  {
     308      chomp;  # eat the trailing newline
     309      # see the Camel book page 266 for hashes of arrays
     310      push @{ $names_by_size{-s $_} }, $_;
     311      # $filecount++;  now handled in next loop
     312  }
     313  
     314  close (pipefd);
     315  
     316  
     317  # delete all entries in names_by_size{} with <2 filenames
     318  # This is done so the progress statistics are more useful.
     319  # The main loop skips all entries with <2 names,
     320  # but this way status can say "processing entry 2 of 200" instead of
     321  # "processing entry 2 of 500" only to find 300 entries are only 1 file.
     322  
     323  # # print "BEFORE: ", scalar (keys %names_by_size), " entries $filecount files\n";
     324  $filecount = 0;
     325  # jsj: DELETE THE SORT
     326  foreach $filesz (sort { $b <=> $a } keys %names_by_size)
     327  {
     328      my $num_files_this_size = scalar (@{$names_by_size{$filesz}});
     329      if ($num_files_this_size < 2)
     330      {
     331  	# # print "... 1 file, sz $filesz, name \"@{$names_by_size{$filesz}}\"\n";
     332  	delete $names_by_size{$filesz};
     333      }
     334      else
     335      {
     336  	$filecount += $num_files_this_size;
     337      }
     338  }
     339  # # print "AFTER : ", scalar (keys %names_by_size), " entries $filecount files\n";
     340  
     341  # # END NEW
     342  
     343  my $num_unique_sizes = scalar (keys %names_by_size);
     344  
     345  print STDERR
     346      "jsjv: found $num_unique_sizes unique file sizes, $filecount files\n"
     347      if ($verbose1);
     348  
     349  # Examine the list of filenames by file size,
     350  # sorting the keys numerically, largest file sizes first.
     351  
     352  my $size_list_index=0;
     353  my $filesz;
     354  foreach $filesz (sort { $b <=> $a } keys %names_by_size)
     355  {
     356      my ($array_ref);
     357      # my (@fc_LoL);  # file compare List of Lists
     358      $size_list_index++;
     359      my $num_files_this_size = scalar (@{$names_by_size{$filesz}});
     360  
     361      # FUTURE DIRECTIONS: add some graphic progress indicator,
     362      # showing progress by number of entries and/or number of files
     363  
     364      if (!defined ($filesz))  # this should never happen, but handle it anyway
     365      {
     366  	print "$num_files_this_size files of UNDEF:\n    ",
     367  	    join ("\n    ", @{$names_by_size{$filesz}}), "\n";
     368  	next;
     369      }
     370  
     371      if ($filesz == 0) # skip files of zero length
     372      {
     373  	print "$num_files_this_size files of size 0:\n    ",
     374  	    join ("\n    ", @{$names_by_size{$filesz}}), "\n";
     375  	next;
     376      }
     377  
     378      # skip if only 1 file of that size
     379      if ($num_files_this_size < 2)
     380      {
     381  print STDERR "jsjv: skipping entry $size_list_index of $num_unique_sizes : ",
     382  "1 file of size $filesz\n"
     383  if ($verbose1);
     384  
     385  	next;
     386      }
     387  
     388  print STDERR "jsjv: processing entry $size_list_index of $num_unique_sizes : ",
     389  "$num_files_this_size files of size $filesz\n" if ($verbose1);
     390  
     391      if ($filesz > 1024)
     392      {
     393  	my (@name_list) = compare_by_first_block (1024,
     394  		@{$names_by_size{$filesz}});
     395  
     396  	# for each list of files of the same size,
     397  	# divide into a list-of-lists by comparing the first block.
     398  
     399  
     400  	# jsj: THIS NEEDS REFINEMENT:
     401  	# the MD5 algorithm saves time under SOME circumstances.
     402  	# Can this be determined using only $filesz and $num_files_this_size ?
     403  	#
     404  	# The underlying tradeoff: md5 reads each file ONCE,
     405  	# but all files with matching checksums
     406  	# must be read again to verify that they are truly identical.
     407  	#
     408  	# compare_by_fc() (compare all files using file compare)
     409  	# used alone saves time when either
     410  	# a) many files match entirely
     411  	# b) the files don't match near the beginning
     412  	# *BUT* it takes SIGNIFICANTLY LONGER for the case of
     413  	# many files if they don't match near the beginning.
     414  	#
     415  	# The first pass: comparing by first block only seems a reasonable
     416  	# compromise and performance enhancement
     417  	# by grouping files that match/differ based on the first block.
     418  	# And all files' first blocks are read into a buffer
     419  	# so no need to re-read for every compare!
     420  
     421  	# # my @md5_LoL;
     422  
     423  	if ($num_files_this_size > 20)
     424  	{
     425  	    for $array_ref (@name_list)
     426  	    {
     427  		compare_by_md5_sum (@$array_ref);  # results in @md5_LoL
     428  	    }
     429  	}
     430  	else
     431  	{
     432  	    @md5_LoL = @name_list;  # skip md5 sum step
     433  	}
     434  
     435  	# note: shift() is used to consume the list of lists
     436  	# If md5_LoL is needed after this point, use foreach() here
     437  	# but be sure to delete/free the list
     438  	# before the main loop iterates again
     439  	# so new results are not appended to the previous loop!
     440  
     441  	while ($array_ref = shift (@md5_LoL))
     442  	{
     443  	    # print "\t    list: ", scalar (@$array_ref), " files\n";
     444  	    # print "\t\t", join ("\n\t\t", @$array_ref), "\n";
     445  	    compare_by_fc (@$array_ref);  # results in @fc_LoL
     446  	}
     447      }
     448      else
     449      {
     450  	# For files small enough to be read entirely into RAM,
     451  	# handle them directly on one pass!
     452  
     453  	@fc_LoL = compare_by_first_block ($filesz, @{$names_by_size{$filesz}});
     454      }
     455  
     456      # the global list of lists fc_LoL has the results from either case
     457      while ($array_ref = shift (@fc_LoL))
     458      {
     459  	print scalar (@$array_ref) .
     460  	    " identical files of size $filesz:\n";
     461  	test_hard_links (@$array_ref);
     462      }
     463  }
     464  
     465  
     466  
     467  # compare_by_first_block: given a list of file names of same sized files,
     468  #	read just the first block and compare files based on that
     469  #	to generate lists of likely matches.
     470  
     471  # args: buffer size, list of files
     472  # returns: array of array of filenames that match by first block
     473  sub compare_by_first_block
     474  {
     475      my ($readsize) = shift (@_);
     476      my (@filenames) = @_;
     477  
     478      my ($first_block);         # read a block here
     479      my (%first_block_by_name); # save block by filename
     480  	# this can get rather large, thus the need to free this ASAP
     481  	# and return only the filenames as results
     482  
     483      my ($filenm);
     484      my ($first_file);
     485      my (@return_LoL);
     486  
     487  # print "phase 2 input: ", scalar (@filenames), " names, buffer size=$readsize\n"; # jsj: verbose
     488  # print "\t", join ("\n\t", @filenames), "\n";
     489  
     490  
     491      #  pass 1: for each file, hash the first block by filename
     492      #      $first_block_by_name { $filenm } = $first_block;
     493      #  Files that cannot be opened or read entirely
     494      #  are NOT on the resulting list.
     495  
     496      foreach $filenm (@filenames)
     497      {
     498  	# print "\ttrying file $filenm\n";
     499  
     500  	# jsj: use BINMODE for non-Unix systems
     501  	unless (open FD, "<$filenm")
     502  	{
     503  	    print STDERR "ERROR: cannot open file \"$filenm\", skipping\n";
     504  	    next;
     505  	}
     506  
     507  # turn off UTF-8 Unicode
     508  # jsjj binmode(FD, ":slurp :raw");
     509  binmode(FD, ":raw");
     510  
     511  	my ($kount) = read FD, $first_block, $readsize;
     512  	if ($kount != $readsize)
     513  	{
     514  	    print STDERR "ERROR: file \"$filenm\", is short ",
     515  			 "(read $kount of $readsize bytes), skipping\n";
     516  	}
     517  	else
     518  	{
     519  	    $first_block_by_name { $filenm } = $first_block;
     520  	}
     521  
     522  	unless (close FD)
     523  	{
     524  	    print STDERR "WARNING: cannot close file \"$filenm\"\n";
     525  	}
     526  
     527      }
     528  
     529      #  pass 2: compare all files to each other
     530      #  There's probably a name for this algorithm:
     531      #
     532      # ex: files a, b, c, d, e
     533      # if none match, the following files are compared
     534      #	(a,b) (a,c) (a,d) (a,e)
     535      #	(b,c) (b,d) (b,e)
     536      #	(c,d) (c,e)
     537      #	(d,e)
     538      # BUT if any match, delete them from the list
     539      #
     540      # ex: in the list (a,b,c,d,e), files a, c, e are all identical
     541      # so in the first pass, (a,c) and (a,e) comparisons succeed
     542      # leaving the file list as b, d
     543      # since there's no need to further compare files that matched!
     544      #
     545      # Oh yea, and save all the successes as an array of arrays.
     546  
     547      my (@file_list) = sort keys %first_block_by_name;
     548  # print "file list = ", join ",", @file_list, "\n";
     549  
     550      while ($first_file = shift (@file_list))
     551      {
     552  	my (@dups_found) = $first_file;
     553  	my (@not_dups);
     554  
     555  	foreach $filenm (@file_list)
     556  	{
     557  # print "\tcompare $first_file to $filenm\n";
     558  	    if ($first_block_by_name {$first_file} eq
     559  		$first_block_by_name {$filenm})
     560  	    {
     561  		push @dups_found, $filenm;
     562  	    }
     563  	    else
     564  	    {
     565  		push @not_dups, $filenm;
     566  	    }
     567  
     568  	}
     569  
     570  # print "\tdups_found = ", join ",", @dups_found, "\n";
     571  # print "\tnot_dups = ", join ",", @not_dups, "\n";
     572  	push @return_LoL , [ @dups_found ]
     573  	    if (scalar (@dups_found) > 1);
     574  	    # save results only if >1 matched
     575  	    # since the first file name is always pushed
     576  
     577  	@file_list = @not_dups;  # continue searching the non-matched files
     578      }
     579  return @return_LoL;
     580  }
     581  
     582  
     583  
     584  # compare_by_md5_sum(): given a list of file names of same sized files,
     585  #   compute the md5sum for each file,
     586  #   generating lists of files with the same sum
     587  #   (which need further verification before declaring they're identical).
     588  #   A null list is returned if nothing matches by md5.
     589  #
     590  # args: list of same sized files
     591  # returns: nothing directly
     592  # GLOBALS USED:
     593  #   appends the lists of filenames that match by md5sum
     594  #   to @md5_LoL, thus gathering the results to all calls allowing
     595  #	for $array_ref ( @name_list )
     596  #	{
     597  #	    compare_by_md5_sum (@$array_ref);
     598  #	}
     599  #	while ($array_ref = shift (@md5_LoL))
     600  #	{
     601  #	    ...
     602  #
     603  
     604  sub compare_by_md5_sum
     605  {
     606      my (%file_name_by_md5sum);  # gather file names by hash result
     607      my ($md5sum);
     608      my ($filename);
     609      my ($md5) = Digest::MD5->new;
     610      # # my (@md5_LoL);
     611  
     612  print STDERR "jsjvv: entering md5sum with ", scalar (@_), " filenames\n"
     613      if ($verbose1);
     614  
     615      # for each file, compute the MD5 digest sum,
     616      # gathering file names of the same hash via a hash of arrays
     617      # (the hash index is the checksum, pointing to a list of filenames
     618      # that gave that same result)
     619  
     620      foreach $filename (@_)
     621      {
     622  	unless (open MD5_fd, $filename)
     623  	{
     624  	    print STDERR "ERROR: cannot open file \"$filename\", skipping\n";
     625  	    next;
     626  	}
     627  
     628  	$md5->reset();
     629  	$md5->addfile(MD5_fd);
     630  	# OLDmd5 $md5sum = $md5->hexdigest();
     631  	$md5sum = $md5->digest();
     632  	close(MD5_fd);
     633  
     634  # jsjnew  $digest = $ctx->digest;
     635  # jsjnew  $digest = $ctx->hexdigest;
     636  
     637  	push @{ $file_name_by_md5sum {$md5sum} }, $filename;
     638      }
     639  
     640      #  Append all lists of >1 files to the global list-of-lists
     641      #  (to allow calling this in a loop and handling all the results
     642      #   after the loop)
     643  
     644      foreach $md5sum (keys %file_name_by_md5sum)
     645      {
     646  	push @md5_LoL , [ @{$file_name_by_md5sum{$md5sum}} ]
     647  	    if ( scalar  (@{$file_name_by_md5sum{$md5sum}}) > 1);
     648  	    # do NOT push single file lists, there's nothing to match!
     649  
     650  print STDERR "jsjvv: md5sum found " .
     651  	    scalar  (@{$file_name_by_md5sum{$md5sum}}) . " matches\n"
     652      if ($verbose1) && ( scalar  (@{$file_name_by_md5sum{$md5sum}}) > 1);
     653      }
     654  }
     655  
     656  
     657  
     658  # compare_by_fc: compare the list of files using file compare
     659  # arg: list of filenames
     660  # GLOBALS USED:
     661  #	@fc_LoL : list of lists of matching files
     662  #	results are appended thus allowing calling this in a loop
     663  # 	and gathering the results for use afterwards.
     664  
     665  sub compare_by_fc
     666  {
     667      my (@file_list) = @_;
     668      my ($first_file);
     669  
     670      while ($first_file = shift (@file_list))
     671      {
     672  	my (@dups_found) = $first_file;
     673  	my (@not_dups);
     674  	my ($filenm);
     675  
     676  	foreach $filenm (@file_list)
     677  	{
     678  	    if (compare ($first_file, $filenm) == 0)
     679  	    {
     680  		push @dups_found, $filenm;
     681  	    }
     682  	    else
     683  	    {
     684  		push @not_dups, $filenm;
     685  	    }
     686  
     687  	}
     688  
     689  	push @fc_LoL , [ @dups_found ]
     690  	    if (scalar (@dups_found) > 1);
     691  	    # save results only if >1 matched
     692  	    # since the first file name is always pushed
     693  
     694  	@file_list = @not_dups;  # continue searching the non-matched files
     695      }
     696  }
     697  
     698  
     699  
     700  # test_hard_links():
     701  # args: list of filenames of identical files
     702  # output: to stdout
     703  
     704  # Within duplicate files, list those that are hard links.
     705  # That's important to know since links are just multiple names
     706  # for the same file and do NOT take up additional disk space!
     707  # (well, except for the additional directory entry).
     708  #
     709  # This is probably Unix specific.
     710  # The Unix file system allows multiple names for the same file.
     711  # This depends on stat() returning a unique inode and device number per file.
     712  # jsj: test over NFS, Cygwin, ...
     713  #
     714  #
     715  # There are several ways to have multiple names for the same file:
     716  # hard and soft links.  Hard links can be detected since they're limited
     717  # to the same file system
     718  # (they're merely directory entries to the same inode).
     719  # Soft links can be anywhere (other file systems, even on offline NFS machines)
     720  # so there's no way to find them all.
     721  # jsj: are these Win-98 "shortcuts"  What of NT?
     722  #
     723  # ex: input of
     724  #	./gatekeeper.dec.com/bison-1.28.tar.gz
     725  #	./gate1/bison-1.28.tar.gz
     726  #	./gate2/bison-1.28.tar.gz
     727  #	./junk/gatekeeper.dec.com/bison-1.28.tar.gz
     728  # generates the output:
     729  # 4 identical files of size 420341:
     730  #    ./gatekeeper.dec.com/bison-1.28.tar.gz
     731  #    2 are LINKED files:
     732  #	./gate1/bison-1.28.tar.gz
     733  #	./gate2/bison-1.28.tar.gz
     734  #    ./junk/gatekeeper.dec.com/bison-1.28.tar.gz
     735  #
     736  # this means there are actually 3 copies of the file,
     737  # one having 2 names.
     738  
     739  sub test_hard_links
     740  {
     741      my (@file_list) = @_;
     742      my ($first_file);
     743      my ($filenm);
     744  
     745      while ($first_file = shift (@file_list))
     746      {
     747  	my (@dups_found) = $first_file;
     748  	my (@not_dups);
     749  # jsj handle stat failure
     750  	my ($dev1, $ino1) = stat ($first_file);
     751  
     752  	foreach $filenm (@file_list)
     753  	{
     754  	    my ($dev2, $ino2) = stat ($filenm);
     755  	    if (($dev1 == $dev2) && ($ino1 == $ino2))
     756  	    {
     757  		push @dups_found, $filenm;
     758  	    }
     759  	    else
     760  	    {
     761  		push @not_dups, $filenm;
     762  	    }
     763  
     764  	}
     765  
     766  	if (scalar (@dups_found) > 1)
     767  	{
     768  	    print "    ", scalar (@dups_found), " are LINKED files:\n\t",
     769  		join ("\n\t", @dups_found), "\n";
     770  	}
     771  	else
     772  	{
     773  	    print "    $first_file\n";
     774  	}
     775  
     776  	@file_list = @not_dups;  # continue searching the non-matched files
     777      }
     778  }
     779  
     780  # jsj: PRUNE sample:
     781  #	dup6 . -name .netscape -prune -o
     782  
     783  # EXAMPLE: to find all duplicates of a specific file,
     784  # find all files of that size
     785  #	ls -l this_file
     786  #	dup6 . -size zzzzz
     787  #
     788  # ??? add an option like
     789  #	--of-this-file <filename>
     790  #	lists duplicates of that file only
     791  #	(to be precise, files the same size of that file)
     792  
     793  
     794  
     795  
     796  # AND: show "lnn" as a comment, describe how I use them together
     797  
     798  
     799  # 3/25/2005
     800  # IDEA: sort out hard links before all other passes:
     801  #	a) keep FIRST filename in list, use as HASH to find others
     802  #	b) use indirection for ALL FILENAMES:
     803  #	   instead of using filename, use dev/ino # to hash of list-of-names
     804  #
     805  # look at the source: how does "du" filter out hard links?
     806  #
     807  #
     808  # XREF: my WWW of STUPID LINUX TRICKS
     809  # - how file modes are OVERLOADED, sample code for set-gid / manditory file locking
     810  # - find / du / ls / ncheck
     811  #
     812  # HARD and SOFT LINKS:
     813  # And now time for STUPID UNIX FILE SYSTEM TRICKS!
     814  # There are several ways to give one file several names.
     815  #
     816  # First: why have many names for a file?
     817  # Sometimes it's useful, such as an icon appearing in many web pages
     818  # and required in many sub-directories.
     819  # Or a header file that's used in many different places.
     820  # But give something too many names and it's hard to adminster and track!
     821  #
     822  # SYMBOLIC LINKS are similar to Window's "shortcuts":
     823  # the file contains THE NAME of the file (or directory) that's desired.
     824  # A common use is to switch to different things at different times.
     825  # for example:
     826  #	"latest_code" is a symbolic link to "archive/version_5"
     827  #	but when version 6 is ready, just point
     828  #	"latest_code" "archive/version_6" and everything that uses it
     829  #	automatically uses the latest version!
     830  #
     831  # SYMBOLIC LINK GOTCHAS:
     832  # - since it's just the name of the file, the file may not exist or have been re-named,
     833  # resulting in a "file not found".  You need to follow the symbolic links
     834  # to find what's missing!
     835  # - symbolic links may point to any file (even on other file systems!)
     836  # but may easily cause loops, cycles and unuseable paths, particularly if
     837  # multiple symbolic links are traversed on the way to the file!
     838  # There's no protection or enforcement of the file-tree-structure.
     839  # - there's no deterministic way to backtrack to find the symbolic links TO A FILE.
     840  # (you can't ask "what points to me" because the symbolic links may be ANYWHERE
     841  # and there are NO back-pointers).
     842  #
     843  #
     844  # HARD LINKS have been around since Unix began.
     845  # It's a way to give non-directory (leaf-node) files more than one name.
     846  #
     847  # [start tech stuff]
     848  # In the Unix File system, directories are merely a list of file names
     849  # and their inode numbers.  An "inode number" points to an "inode": the structure
     850  # that REALLY describes a file (file type, mode, owner, disk blocks, timestamps).
     851  # It's perfectly valid for one inode to have more than one directory entry
     852  # point to it (that's naturally limited to within a file system
     853  # because inode numbers are only unique within each file system.
     854  # Every Unix/Linux file system starts at inode 1).
     855  # That's why
     856  # - the low level operation to remove a file name is "unlink"
     857  # - the low level operation to attach a name to a file is "link",
     858  #   available as the "ln" shell command
     859  # - The second column of "ls -l" shows the LINK COUNT: the number of names for a file.
     860  # [end tech stuff]
     861  #
     862  # To demonstrate, create a file
     863  #	touch real_file
     864  #	ls -li
     865  # 312832 -rw-rw-r--    1 jeffj    jeffj           0 Mar 25 21:05 real_file
     866  # ^^ inode number      ^ link count: there's one name for this file
     867  #
     868  # now create a soft/symbolic link to the file
     869  #	ln -s real_file sym_link
     870  # and create a hard link to the file
     871  #	ln real_file hard_link
     872  # and list them
     873  #	ls -li
     874  #	312832 -rw-rw-r--    2 jeffj    jeffj           0 Mar 25 21:05 hard_link
     875  #	312832 -rw-rw-r--    2 jeffj    jeffj           0 Mar 25 21:05 real_file
     876  #	312834 lrwxrwxrwx    1 jeffj    jeffj           9 Mar 25 21:05 sym_link -> real_file
     877  #       ^ inode number       ^ link count
     878  #
     879  # "real_file" and "hard_link" are both names for inode 312832:
     880  # that's why both show a "2" for link count.
     881  # Both names are equally valid for the same file.
     882  # Edit/change/alter one file and the other's the same contents since it's ONE PHYSICAL FILE!
     883  #
     884  # "sym_link" is different file, that's why it has its own inode number
     885  # and a link count of 1.  It points to the file named "real_file".
     886  #
     887  # rename "real_file" to "backup"
     888  #	mv real_file backup
     889  # - sym_file is now INVALID!  There's no file named "real_file".
     890  #	("ls --color=tty" flashes it in red to indicate that)
     891  # - but the file still has 2 names.  The inode (312832) is unchanged
     892  #	because only the directory names were changed.
     893  #
     894  #	312832 -rw-rw-r--    2 jeffj    jeffj           0 Mar 25 21:05 backup
     895  #	312832 -rw-rw-r--    2 jeffj    jeffj           0 Mar 25 21:05 hard_link
     896  #	312834 lrwxrwxrwx    1 jeffj    jeffj           9 Mar 25 21:05 sym_link -> real_file
     897  #
     898  # delete "backup" and the file remains as "hard_link"
     899  #	312834 lrwxrwxrwx    1 jeffj    jeffj           9 Mar 25 21:05 sym_link -> real_file
     900  #	312832 -rw-rw-r--    1 jeffj    jeffj           0 Mar 25 21:05 hard_link
     901  #       ^ same file inode    ^ only 1 name points to this file now
     902  #
     903  # zzz tell how I use "lnn" to link duplicates once found with this
     904  #	particularly how the "already linked" file MUST BE FIRST:
     905  # show before/after of ls -li
     906  # zzz AND SRC OF LNN
     907  
